/* aeconfig.h -- Aexpl config file handling
 *
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Enrico Kern <phantom@h07.org>
 *          Christian Volmering <risk@h07.org>
 *
 * This file is part of the aexpl-ng distribution.
 *
 * The aexpl-ng distribution is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef _H07_AECONFIG_H
#define _H07_AECONFIG_H

#include <limits.h>

#define CFG_VALUE_STR(decl) static decl = {'\0'}
#define CFG_VALUE_INT(decl) static decl = 0;
#define CFG_VALUE_BOOL(decl) CFG_VALUE_INT(decl)
#define CFG_VALUE_CB(decl) static decl = {0, {'\0'}, {'\0'}, {'\0'}}

enum _cfg_etype_t
{
  etype_none = 0,
  etype_str,
  etype_int,
  etype_bool,
  etype_custom_block
};

struct _cfg_element_t
{
  const char *path;
  void *buf;
  enum _cfg_etype_t type;
  unsigned short optional;
  int size;
};

struct _cfg_custom_block_t
{
  unsigned short available;
  char exploit_db[PATH_MAX];
  char update_url[PATH_MAX];
  char ca_cert[PATH_MAX];
};

typedef struct _cfg_element_t cfg_element_t;
typedef struct _cfg_custom_block_t cfg_custom_block_t;

extern int cfg_setup (const char *);
extern void cfg_teardown (void);
extern char *cfg_read_str (const char *);
extern int cfg_read_int (const char *);
extern int cfg_read_bool (const char *);
extern cfg_custom_block_t *cfg_read_custom_block (const char *);
extern int cfg_read_elements (const cfg_element_t *, char *, int);

#endif /* _H07_AECONFIG_H */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
