/* main.c -- AntiExploit
 * 
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Enrico Kern <phantom@h07.org>
 *          Christian Volmering <risk@h07.org>
 * 
 * This file is part of the aexpl-ng distribution.
 * 
 * The aexpl-ng distribution is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

#include "aeconfig.h"
#include "loaddb.h"
#include "dazukoio.h"
#include "queue.h"
#include "hashtable.h"
#include "md5sum.h"
#include "sendmail.h"
#include "alert.h"
#include "daemon.h"
#include "proctitle.h"
#include "lock.h"
#include "config.h"
#include "fqueue.h"
#include "log.h"
#include "proxy.h"
#include "version.h"
#include "curl_helpers.h"

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <sys/wait.h>

#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

#if !defined CLD_EXITED
#  define CLD_EXITED 1		// Child has exited.
#endif
#if !defined CLD_KILLED
#  define CLD_KILLED 2		// Child was killed.
#endif
#if !defined CLD_DUMPED
#  define CLD_DUMPED 3		// Child terminated abnormally.
#endif

static unsigned short running = 1;
static htable_t aexpl_db = NULL;
static htable_t custom_db = NULL;
static fqueue_t filequeue = NULL;
static pid_t worker_th = -1;
static pid_t main_th = -1;

static void sigterm_handler (int sig);
static void sighup_handler (int sig);
static void sigpipe_handler (int sig);
static void sigchld_handler (int sig, siginfo_t * info, void *context);
static int register_dazuko (void);
static void worker_loop (void);
static void main_loop (void);
static int configure (const char *config);
static int initialise (void);
static void exit_cleanup (void);
static void opt_version (void);
static void opt_usage (const char *self);
static void opt_update (const char *config);
static void opt_start (const char *config, int argc, char **argv);
static void opt_stop (const char *config);
static void opt_check (const char *config);

/* {{{ Configuration section */

// Path to exploit database
CFG_VALUE_STR (char exploit_db[PATH_MAX]);
// Uri where the exploit database is downloaded from
CFG_VALUE_STR (char update_url[PATH_MAX]);
// Mail adress for exploit alerts
CFG_VALUE_STR (char mail_addr[128]);
// User script executed on alterts
CFG_VALUE_STR (char user_script[PATH_MAX]);
// Lock file for aexpl (stores pid)
CFG_VALUE_STR (char lock_file[PATH_MAX]);
// Path to the shared library config file
CFG_VALUE_STR (char ld_so_conf[PATH_MAX]);
// Log file for aexpl
CFG_VALUE_STR (char log_file[PATH_MAX]);
// Whether to run the user script or not
CFG_VALUE_BOOL (short run_script);
// Size of inode table
CFG_VALUE_INT (int inot_size);
// Maximum file size to check
CFG_VALUE_INT (int max_file_size);
// Size of the file queue
CFG_VALUE_INT (int queue_size);
// Log facility for syslog
CFG_VALUE_INT (int facility);
// Proxy configuration
CFG_VALUE_CB (proxy_config_t proxy_config);
// Location of CA cert
CFG_VALUE_STR (char ca_cert[PATH_MAX]);
// Location of online version file
CFG_VALUE_STR (char version_url[PATH_MAX]);
// Custom database
CFG_VALUE_CB (cfg_custom_block_t custom_block);
// Exclude paths
CFG_VALUE_STR (char exclude_paths[PATH_MAX]);

static cfg_element_t ae_config[] = {
  {"/aexpl/analyze/exploit_db", exploit_db, etype_str, 0, PATH_MAX},
  {"/aexpl/analyze/update_url", update_url, etype_str, 0, PATH_MAX},
  {"/aexpl/reports/mail_addr", mail_addr, etype_str, 0, 128},
  {"/aexpl/reports/log_file", log_file, etype_str, 1, PATH_MAX},
  {"/aexpl/scripts/user_script", user_script, etype_str, 1, PATH_MAX},
  {"/aexpl/core/lock_file", lock_file, etype_str, 0, PATH_MAX},
  {"/aexpl/core/ld_so_conf", ld_so_conf, etype_str, 0, PATH_MAX},
  {"/aexpl/scripts/run_script", &run_script, etype_bool, 1},
  {"/aexpl/core/inot_size", &inot_size, etype_int, 0},
  {"/aexpl/core/max_file_size", &max_file_size, etype_int, 0},
  {"/aexpl/core/queue_size", &queue_size, etype_int, 0},
  {"/aexpl/reports/facility", &facility, etype_int, 0},
  {"/aexpl/proxy/enabled", &proxy_config.enabled, etype_bool, 1},
  {"/aexpl/proxy/type", proxy_config.type, etype_str, 1, 7},
  {"/aexpl/proxy/host", proxy_config.host, etype_str, 1, 256},
  {"/aexpl/proxy/login", proxy_config.login, etype_str, 1, 128},
  {"/aexpl/analyze/ca_cert", ca_cert, etype_str, 1, PATH_MAX},
  {"/aexpl/analyze/version_url", version_url, etype_str, 0, PATH_MAX},
  {"/aexpl/analyze/custom_db", &custom_block, etype_custom_block, 1},
  {"/aexpl/analyze/exclude_paths", &exclude_paths, etype_str, 1, PATH_MAX},
  {0, 0, etype_none, 0, 0}
};

/* }}} */

/* {{{ main */
int
main (int argc, char **argv)
{
  char *config = NULL;

#if !defined(AEXPL_CONF)
  static char optstring[] = "c:u:t:o:vh";
#else
  static char optstring[] = "cutovh";
  config = AEXPL_CONF;
#endif

  if (!argc || !config)
    opt_usage (*argv);

  switch (getopt (argc, argv, optstring))
    {
    case 'v':
      opt_version ();
      break;

    case 'h':
      opt_usage (*argv);
      break;

    case 'u':
      opt_update (config);
      break;

    case 'c':
      opt_start (config, argc, argv);
      break;

    case 't':
      opt_stop (config);
      break;

    case 'o':
      opt_check (config);
      break;

    default:
      opt_usage (*argv);
    }

  return (0);
}

/* }}} */

/* {{{ opt_version */
static void
opt_version (void)
{
  printf ("AntiExploit %s\n", VERSION);
  printf ("Copyright (C) 2004 H Zero Seven\n\n");

  printf ("This program comes with ABSOLUTELY NO WARRANTY.\n");
  printf ("This is free software, and you are welcome to "
	  "redistribute it\nunder certain conditions. "
	  "See the file COPYING for details.\n");

  exit (0);
}

/* }}} */

/* {{{ opt_usage */
static void
opt_usage (const char *self)
{
  printf ("Usage: %s [options] [config]\nOptions:\n", self);
  printf ("  %-4s %s\n", "-h", "Print this message and exit.");
  printf ("  %-4s %s\n", "-v", "Show version information.");
  printf ("  %-4s %s\n", "-c", "Start AntiExploit.");
  printf ("  %-4s %s\n", "-t", "Stop AntiExploit.");
  printf ("  %s * %s\n", "-u", "Update the Exploit database.");
  printf ("  %s * %s\n", "-o", "Check for updated versions.");
  printf ("\n* This operation will open a remote connection!\n");
  printf ("\nReport bugs to %s.\n", "phantom@h07.org or risk@h07.org");

  exit (0);
}

/* }}} */

/* {{{ opt_start */
static void
opt_start (const char *config, int argc, char **argv)
{
  struct sigaction sa;

  if (getuid () != 0)
    {
      printf ("Only root can run AntiExploit!\n");

      exit (1);
    }

  atexit (exit_cleanup);

  if (!configure (config))
    exit (1);

  if (read_pid_file (lock_file))
    {
      fprintf (stderr,
	       "Another instance of AntiExploit is already running!\n");

      exit (1);
    }

  printf ("Detaching, please check the log file to"
	  " see if everything is ok!\n");

  if (!detach (1, 0))
    exit (1);

  if (!initialise ())
    exit (1);

  // create file queue
  if (!(filequeue = new_fqueue (queue_size)))
    {
      write_log (LOG_ERR, "Error creating file queue: %s", strerror (errno));

      exit (1);
    }

  if (!create_pid_file (lock_file))
    {
      write_log (LOG_ERR, "Error creating lock file: %s", strerror (errno));

      exit (1);
    }

  write_log (LOG_INFO, "AntiExploit started.");

  set_proc_title (argc, argv, "[aexpl]");

  sigemptyset (&sa.sa_mask);
  sa.sa_flags = SA_RESTART;

  sa.sa_handler = sigterm_handler;
  if (sigaction (SIGTERM, &sa, NULL) == -1)
    {
      write_log (LOG_ERR, "Error setting up signal handler: %s",
		 strerror (errno));

      exit (1);
    }

  sa.sa_handler = sighup_handler;
  if (sigaction (SIGHUP, &sa, NULL) == -1)
    {
      write_log (LOG_ERR, "Error setting up signal handler: %s",
		 strerror (errno));

      exit (1);
    }

  sa.sa_handler = sigpipe_handler;
  if (sigaction (SIGPIPE, &sa, NULL) == -1)
    {
      write_log (LOG_ERR, "Error setting up signal handler: %s",
		 strerror (errno));

      exit (1);
    }

  sa.sa_handler = NULL;
  sa.sa_sigaction = sigchld_handler;
  sa.sa_flags = SA_RESTART | SA_SIGINFO;
  if (sigaction (SIGCHLD, &sa, NULL) == -1)
    {
      write_log (LOG_ERR, "Error setting up signal handler: %s",
		 strerror (errno));

      exit (1);
    }

  main_th = getpid ();
  switch (worker_th = fork ())
    {
    case 0:
      worker_loop ();
      break;

    case -1:
      write_log (LOG_ERR, "Couldn't fork: %s", strerror (errno));
      break;

    default:
      main_loop ();
    }

  remove_pid_file (lock_file);

  exit (0);
}

/* }}} */

/* {{{ opt_update */
static void
opt_update (const char *config)
{
  int ae_pid;

  if (!configure (config))
    exit (1);

  printf ("Updating exploit database: ");
  if (!curl_get (update_url, exploit_db, NULL, 0, proxy_config, ca_cert))
    {
      printf ("failed!\n");

      exit (1);
    }
  else
    printf ("done.\n");

  if (custom_block.available)
    {
      printf ("Updating custom database: ");
      if (!curl_get (custom_block.update_url, custom_block.exploit_db,
		     NULL, 0, proxy_config, custom_block.ca_cert))
	printf ("failed!\n");
      else
	printf ("done.\n");
    }

  if ((ae_pid = read_pid_file (lock_file)))
    {
      printf ("\nSending SIGHUP to active AntiExploit process");

      if (kill (ae_pid, SIGHUP) != 0)
	{
	  perror ("");

	  exit (1);
	}
      else
	printf (": done.\n");
    }

  exit (0);
}

/* }}} */

/* {{{ opt_stop */
static void
opt_stop (const char *config)
{
  int ae_pid;

  if (!configure (config))
    exit (1);

  if ((ae_pid = read_pid_file (lock_file)))
    {
      printf ("Sending SIGTERM to active AntiExploit process: ");

      if (kill (ae_pid, SIGTERM) != 0)
	{
	  perror ("Couldn't send signal");

	  exit (1);
	}
      else
	printf ("done.\n");
    }
  else
    printf ("AntiExploit process not running!\n");

  exit (0);
}

/* }}} */

/* {{{ opt_check */
static void
opt_check (const char *config)
{
  char online_version[42];
  ae_version_t v_current = VERSION_INIT;
  ae_version_t v_online = VERSION_INIT;

  if (!configure (config))
    exit (1);

  if (!curl_get
      (version_url, NULL, online_version, 42, proxy_config, ca_cert))
    exit (1);

  if (!parse_version_str (VERSION, &v_current)
      || !parse_version_str (online_version, &v_online))
    {
      printf ("Invalid version format!\n");
      exit (1);
    }

  if (VERSION_CMP (v_online, v_current))
    {
      get_version_str (&v_online, online_version);
      printf ("A new version `%s' of AntiExploit is available.\n",
	      online_version);
    }
  else
    printf ("You're running the latest version of AntiExploit.\n");
}

/* }}} */

/* {{{ sigterm_handler */
static void
sigterm_handler (int sig)
{
  int chld_status;

  write_log (LOG_INFO, "Process %ld terminating ...", (long) getpid ());

  running = 0;

  if (getpid () == main_th)
    {
      if (dazukoUnregister () != 0)
	write_log (LOG_ERR, "Failed to unregister with Dazuko.");

      free_fqueue (filequeue);

      if (worker_th != -1)
	{
	  // Child would send SIGTERM to the main_thread so ignore this
	  signal (SIGCHLD, SIG_IGN);

	  // Kill the worker thread and wait until it is terminated
	  kill (worker_th, SIGTERM);

	  while ((waitpid (worker_th, &chld_status, 0) == -1)
		 && (errno == EINTR));
	}
    }
}

/* }}} */

/* {{{ sighup_handler */
static void
sighup_handler (int sig)
{
  pid_t pid;

  pid = getpid ();

  if (pid == main_th)
    {
      if (worker_th != -1)
	{
	  write_log (LOG_INFO, "Sending SIGHUP to child: %ld",
		     (long) worker_th);

	  kill (worker_th, SIGHUP);
	}
    }
  else
    {
      write_log (LOG_INFO, "Got SIGHUP, updating database.");

      free_htable (aexpl_db);
      if (loaddb (&aexpl_db, exploit_db) < 0)
	{
	  write_log (LOG_ERR, "Error reading exploit database.");
	  running = 0;
	}

      if (custom_block.available)
	{
	  if (custom_db)
	    free_htable (custom_db);

	  if (loaddb (&custom_db, custom_block.exploit_db) < 0)
	    write_log (LOG_ERR, "Error reading custom database.");
	  else
	    write_log (LOG_INFO, "Custom database loaded with %i entries.",
		       htable_entries (custom_db));
	}

      write_log (LOG_INFO, "Exploit database loaded with %i entries.",
		 htable_entries (aexpl_db));
    }
}

/* }}} */

/* {{{ sigpipe_hanlder */
static void
sigpipe_handler (int sig)
{
  short b_result = 1;
  char c_result[32];

  if (getpid () == main_th)
    {
      write_log (LOG_ERR, "Broken pipe in main thread.");

      // Try to reregister with dazuko, the broken pipe is probably from
      // the socket connected to the dazuko module.
      if ((b_result = (!dazukoUnregister () && register_dazuko ())))
	snprintf (c_result, 32, "success");
      else
	snprintf (c_result, 32, "failed");

      write_log (LOG_INFO, "Reregistering with dazuko: %s.", c_result);

      if (!b_result)
	kill (main_th, SIGTERM);
    }
  else
    {
      write_log (LOG_ERR, "Broken pipe in worker thread.");

      kill (main_th, SIGTERM);
    }
}

/* }}} */

/* {{{ sigchld_handler */
static void
sigchld_handler (int sig, siginfo_t * info, void *context)
{
  if (getpid () != main_th)
    return;

  switch (info->si_code)
    {
    case CLD_EXITED:
      kill (main_th, SIGTERM);

      break;

    case CLD_KILLED:
      if (info->si_status)
	write_log (LOG_INFO, "Process %ld terminated due to signal: %d",
		   info->si_pid, info->si_status);
      kill (main_th, SIGTERM);

      break;

    case CLD_DUMPED:
      write_log (LOG_INFO, "Process %d terminated abnormally.", info->si_pid);
      kill (main_th, SIGTERM);

      break;
    }
}

/* }}} */

/* {{{ register_dazuko */
static int
register_dazuko (void)
{
  char tmp[PATH_MAX];
  char *phrase, *last;

  if (dazukoRegister ("ANTI_EXPLOIT", "r") != 0)
    return (0);
  else
    {
      FILE *fp_so_conf;
      char lib_path[PATH_MAX];

      // Add all library paths to exclude
      if ((fp_so_conf = fopen (ld_so_conf, "r")))
	{
	  while (fgets (lib_path, PATH_MAX, fp_so_conf) > 0)
	    dazukoAddExcludePath (lib_path);
	}

      if (dazukoAddIncludePath ("/") != 0 ||
	  dazukoAddExcludePath ("/dev") != 0 ||
	  dazukoAddExcludePath ("/proc") != 0 ||
	  dazukoAddExcludePath ("/lib") != 0)
	{
	  write_log (LOG_ERR, "Could not add include or exclude paths.");

	  return (0);
	}

      if (exclude_paths)
	{
	  strncpy (tmp, exclude_paths, PATH_MAX);
	  for (phrase = strtok_r (tmp, ":", &last);
	       phrase; phrase = strtok_r (NULL, ":", &last))
	    dazukoAddExcludePath (phrase);
	}

      // set access mask
      if (dazukoSetAccessMask (DAZUKO_ON_OPEN |
			       DAZUKO_ON_CLOSE |
			       DAZUKO_ON_CLOSE_MODIFIED |
			       DAZUKO_ON_EXEC) != 0)
	{
	  write_log (LOG_ERR, "Failed to set access mask.");
	  dazukoUnregister ();

	  return (0);
	}

      return (1);
    }
}

/* }}} */

/* {{{ on_dequeue */
static void
on_dequeue (void *sender, const void *ptr, int action)
{
  long *ino_ptr = (long *) ptr;
  htable_t ino_table = (htable_t) sender;

  // When action is dequeue the queue removed the last inode
  // because it's filled more than max_size, so we alsow delete
  // the inode from the hash table.
  if (action == NOTIFY_DEQUEUE)
    {
      htable_delete (ino_table, ino_ptr);
      free (ino_ptr);
    }
}

/* }}} */

/* {{{ worker_loop */
static void
worker_loop (void)
{
  char file[PATH_MAX];
  unsigned long long *ino;
  char sum[33];
  struct stat file_stat;
  htable_t ino_table;
  queue_t ino_queue;

  // This is the md5 sum of a zero length file.
  const char *zero_sum = "d41d8cd98f00b204e9800998ecf8427e";

  write_log (LOG_INFO, "Worker thread woken up.");

  // Try to create the inode table to store allready scanned files.
  if (!(ino_table = new_htable (inot_size, 0, KEY_ULONGLONG)))
    {
      write_log (LOG_ERR, "Error creating inotab: %s", strerror (errno));

      exit (1);
    }

  if (!(ino_queue = new_queue (inot_size)))
    {
      write_log (LOG_ERR, "Error creating inoqueue: %s", strerror (errno));

      exit (1);
    }

  notify (ino_queue, (void *) ino_table, &on_dequeue);

  open_fqueue (filequeue);

  while (running)
    {
      if (!fdequeue (filequeue, file))
	continue;

      if (stat (file, &file_stat) == -1)
	continue;

      // We only want reglular files and again, only files bigger than zero.
      if (!S_ISREG (file_stat.st_mode) || (file_stat.st_size == 0))
	continue;

      if (!(ino = (unsigned long long *)
	    malloc (sizeof (unsigned long long))))
	{
	  write_log (LOG_ERR, "Couldn't allocate inode: %s",
		     strerror (errno));

	  exit (1);
	}

      // A inode is not unique, so we do a bitwise xor with the device id.
      *ino = file_stat.st_dev ^ file_stat.st_ino;

      // If the table contains the inode, skip the checks.
      if (!htable_contains (ino_table, ino))
	{
	  if (!md5sum_from_file (file, sum))
	    free (ino);
	  else
	    {
	      // Check if the sum is from a zero length file and if not,
	      // check if the exploit db contains this md5 sum.
	      if (strncmp (zero_sum, sum, 32))
		{
		  if ((aexpl_db && htable_contains (aexpl_db, sum))
		      || (custom_db && htable_contains (custom_db, sum)))
		    {
		      alert (file, mail_addr,
			     run_script ? user_script : NULL,
			     file_stat.st_uid, file_stat.st_gid);
		    }
		}

	      // Insert inode in the hashtable tp prevend multible checks. 
	      htable_insert (ino_table, (void *) ino, NULL);

	      // Alsow insert into ino queue.
	      //
	      // The queue will remove the last inode if it's filled more
	      // than max_size and the notify callback will than remove
	      // the inode from the hash table.
	      enqueue (ino_queue, (void *) ino);
	    }
	}
      else
	free (ino);
    }

  free_queue (ino_queue);
  free_htable (ino_table);

  exit (0);
}

/* }}} */

/* {{{ main_loop */
static void
main_loop (void)
{
  struct dazuko_access *acc = NULL;
  int fail_count = 0;

  while (running)
    {
      if (acc != NULL)
	dazukoReturnAccess (&acc);

      // Wait for a access event.
      if (dazukoGetAccess (&acc) == 0)
	{
	  fail_count = 0;

	  if (!acc->set_event || !acc->set_file_size)
	    continue;

	  // Exclude files bigger than MAX_FILE_SIZE or zero length.
	  if ((acc->file_size > max_file_size) || (acc->file_size == 0))
	    continue;

	  switch (acc->event)
	    {
	    case DAZUKO_ON_OPEN:
	      fenqueue (filequeue, acc->filename);
	      break;

	    case DAZUKO_ON_CLOSE:
	      fenqueue (filequeue, acc->filename);
	      break;

	    case DAZUKO_ON_CLOSE_MODIFIED:
	      fenqueue (filequeue, acc->filename);
	      break;

	    case DAZUKO_ON_EXEC:
	      fenqueue (filequeue, acc->filename);
	      break;
	    }
	}
      else
	{
	  fail_count++;

	  if (fail_count > 2)
	    {
	      write_log (LOG_ERR, "Error getting access three times,"
			 " terminating.");

	      running = 0;
	    }
	}
    }
}

/* }}} */

/* {{{ configure */
static int
configure (const char *config)
{
  char err_buf[128];

  if (!cfg_setup (config))
    {
      printf ("Configuration error: %s\n", "Syntax error.");

      return (0);
    }

  if (!cfg_read_elements (ae_config, err_buf, 128))
    {
      printf ("Configuration error: %s\n", err_buf);

      return (0);
    }

  return (1);
}

/* }}} */

/* {{{ initialise */
static int
initialise (void)
{

  unsigned short result = 1;

  setup_log ("AntiExploit", (facility != -1) ? facility : LOG_DAEMON,
	     log_file);

  write_log (LOG_INFO, "Starting initialization.");

  // Load the exploit database.
  if (loaddb (&aexpl_db, exploit_db) < 0)
    {
      write_log (LOG_ERR,
		 "Could not load exploit-database, please verify "
		 "that the file you specified in aexpl.conf "
		 "exist, otherwhise check the README!");

      result = 0;
    }
  else
    {
      write_log (LOG_INFO, "Exploit database loaded with %i entries.",
		 htable_entries (aexpl_db));

      if (custom_block.available)
	{
	  if (loaddb (&custom_db, custom_block.exploit_db) < 0)
	    write_log (LOG_ERR, "Could not load custom database!");
	  else
	    write_log (LOG_INFO, "Custom database loaded with %i entries.",
		       htable_entries (custom_db));
	}

      // If exploit database loaded successfully, register with dazuko.
      if (register_dazuko ())
	{
	  write_log (LOG_INFO, "Successfully registered with dazuko.");

	  result = 1;
	}
      else
	{
	  write_log (LOG_ERR,
		     "Could not register with dazuko please "
		     "verify that dazuko is installed and/or "
		     "read the README of aexpl");

	  result = 0;
	}
    }

  if (result)
    write_log (LOG_INFO, "Initialization complete.");
  else
    write_log (LOG_INFO, "Initialization failed!");

  return (result);
}

/* }}} */

/* {{{ exit_cleanup */
static void
exit_cleanup (void)
{
  teardown_log ();
  dazukoUnregister ();
  cfg_teardown ();
}

/* }}} */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
