/* proctitle.c -- Routine for setting the process title
 * 
 * Copyright (C) 2004 H Zero Seven
 *
 * Authors: Christian Volmering <risk@h07.org>
 * 
 * This file is part of the aexpl-ng distribution.
 * 
 * The aexpl-ng distribution is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 of 
 * the License, or (at your option) any later version.
 *
 * The aexpl-ng distribution is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the aexpl-ng distribution; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 * MA 02111-1307 USA
 */

#include "config.h"
#include "proctitle.h"

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>

#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

/* {{{ set_proc_title */
int
set_proc_title (int argc, char **argv, const char *fmt, ...)
{
  char title[64];
  va_list ap;
  int len, i;

#ifndef HAVE_SETPROCTITLE
  extern char **environ;
  static int max_size = -1;
  char *ptr = NULL, *last = *(argv + (argc - 1));
#endif

  va_start (ap, fmt);
  vsnprintf (title, sizeof (title), fmt, ap);
  len = strlen (title);
  va_end (ap);

#ifndef HAVE_SETPROCTITLE
  if (max_size == -1)
    {
      max_size = (last + strlen (last)) - *argv;

      for (i = 0; ptr; ptr = *(environ + i++))
	max_size += strlen (ptr) + 1;
    }

  strncpy (*argv, title, max_size);

  for (ptr = *argv + strlen (title); ptr < *argv + max_size; *ptr++ = '\0');

  return (max_size > len ? len : max_size);
#else
  setproctitle (title);

  return (len);
#endif
}

/* }}} */

/*
 * Local variables:
 * tab-width: 8
 * c-basic-offset: 8
 * End:
 * vim600: noet nonu sw=8 ts=8 fdm=marker
 * vim<600: noet nonu sw=8 ts=8
 */
